from unittest import TestCase
from operators.python_operator import PythonOperator
from models.dag import DAG


class TestOperatorDependencies(TestCase):

    def get_dag(self):

        default_args = {}

        return DAG(
            dag_id="test_dag",
            description="Test DAG",
            default_args=default_args,
            schedule_interval="0 * * * *"
        )

    def get_task(self,*,task_id,dag):

        def test(arg):
            pass

        return PythonOperator(
            dag=dag,
            task_id=task_id,
            description="Test task",
            python_callable=test,
            op_kwargs={'arg': 'great test'},
            requirements=["pandas==1.0.0", "pandas_gbq==0.14.0"]
        )

    def test_upstream_setting(self):

        dag = self.get_dag()
        test_task = self.get_task(task_id='test_task',dag=dag)
        test_task2 = self.get_task(task_id='test_task2',dag=dag)

        test_task.set_upstream(test_task2)

        upstream_tasks = test_task.upstream_tasks
        downstream_tasks = test_task2.downstream_tasks

        self.assertEqual("test_task2", upstream_tasks[0])
        self.assertEqual("test_task", downstream_tasks[0])

    def test_upstream_lshift_setting(self):

        dag = self.get_dag()
        test_task = self.get_task(task_id='test_task',dag=dag)
        test_task2 = self.get_task(task_id='test_task2',dag=dag)

        test_task << test_task2

        upstream_tasks = test_task.upstream_tasks
        downstream_tasks = test_task2.downstream_tasks

        self.assertEqual("test_task2", upstream_tasks[0])
        self.assertEqual("test_task", downstream_tasks[0])

    def test_downstream_setting(self):

        dag = self.get_dag()
        test_task = self.get_task(task_id='test_task',dag=dag)
        test_task2 = self.get_task(task_id='test_task2',dag=dag)

        test_task.set_downstream(test_task2)

        downstream_tasks = test_task.downstream_tasks
        upstream_tasks = test_task2.upstream_tasks

        self.assertEqual("test_task2", downstream_tasks[0])
        self.assertEqual("test_task", upstream_tasks[0])

    def test_downstream_rshift_setting(self):

        dag = self.get_dag()
        test_task = self.get_task(task_id='test_task',dag=dag)
        test_task2 = self.get_task(task_id='test_task2',dag=dag)

        test_task >> test_task2

        downstream_tasks = test_task.downstream_tasks
        upstream_tasks = test_task2.upstream_tasks

        self.assertEqual("test_task2", downstream_tasks[0])
        self.assertEqual("test_task", upstream_tasks[0])




import click
from dotenv import load_dotenv
import os
import yaml
import subprocess
import json
from pathlib import Path
from .cli.generate_terraform import terraform
load_dotenv()

TERRAFORM_DIRECTORY = "terraform_plan"
CONFIG_FILE = "limber.yaml"

@click.group()
def cli():
    """
    Just the main cli
    """
    load_environment_variables()


def load_environment_variables():
    absolute_config_file = os.path.abspath(CONFIG_FILE)

    with open(absolute_config_file) as file:
        yaml_config = yaml.safe_load(file.read())

    os.environ["CLOUD_FUNCTIONS_SERVICE_ACCOUNT_EMAIL"] = yaml_config["cloud"]["cloud_functions_service_account"]

    os.environ["TERRAFORM_ORGANIZATION"] = yaml_config["terraform"]["organization"]
    os.environ["TERRAFORM_WORKSPACE"] = yaml_config["terraform"]["workspace"]


@cli.command("init")
def init():
    """
    Intializes Limber
    """

    # Create a folder for the output
    Path(TERRAFORM_DIRECTORY).mkdir(exist_ok=True)

    # Create initial terraform config there
    absolute_config_file = os.path.abspath(CONFIG_FILE)

    with open(absolute_config_file) as file:
        yaml_config = yaml.safe_load(file.read())

    config = {
        "provider": {
            yaml_config["cloud"]["provider"]:
                {
                    "project": yaml_config["cloud"]["project"],
                    "region": yaml_config["cloud"]["region"]
                }
        },
        "resource": {
            "google_storage_bucket": {
                "bucket": {
                    "name": yaml_config["cloud"]["default_bucket"]
                }
            }
        },
        "terraform": {
            "backend": {
                "remote": {
                    "organization": os.environ["TERRAFORM_ORGANIZATION"],
                    "workspaces": {
                        "name": os.environ["TERRAFORM_WORKSPACE"]
                    }
                }
            }
        }
    }

    provider_config = f"{TERRAFORM_DIRECTORY}/provider.tf.json"
    with open(provider_config,"w") as file:
        file.write(json.dumps(config, indent=4, sort_keys=False))

    print("Limber has now successfully initialized using your configuration")

t = terraform(folder="terraform_plan")

@cli.command("plan")
def plan():
    """
    Create a plan for infra
    """
    t.create_terraform_configuration()


if __name__ == '__main__':
    cli()


from typing import Optional
from datetime import timedelta


class DAG:

    def __init__(self, dag_id:str, description:Optional[str], default_args:dict, schedule_interval:[str,timedelta]):

        self.dag_id = dag_id
        self.description=description
        self.default_args=default_args
        self.schedule_interval=schedule_interval

        pass

    def get_terraform_json(self) -> {}:

        configuration = {
            "resource": {
                "google_pubsub_topic": [{
                    f"dag_{self.dag_id}": {
                        "name": f"dag_{self.dag_id}"
                    }
                }], "google_cloud_scheduler_job": [{
                    f"job_{self.dag_id}": {
                        "name": self.dag_id,
                        "description": self.description,
                        "schedule": self.schedule_interval,
                        "pubsub_target": {
                            "topic_name": "${google_pubsub_topic.dag_" + self.dag_id + ".id}",
                            "data": "dGVzdA=="
                        }
                    }
                }]
            }
        }

        return configuration


import os
from ..models.operator import Operator
import inspect
import shutil
import hashlib
import zipfile
import stat

class PythonOperator(Operator):

    def __init__(self, *, dag, task_id, description, python_callable, op_kwargs, requirements: [], provide_context=False):
        super().__init__()

        self.dag = dag
        self.task_id = task_id
        self.description = description
        self.python_callable = python_callable
        self.op_kwargs = op_kwargs
        self.requirements = requirements
        self.provide_context = provide_context

    def _get_func_parameters(self, kwargs) -> str:

        parameters = []
        for key, value in kwargs.items():
            parameters.append(f"{key}='{value}'")

        if self.provide_context:
            parameters.append("context=data")

        return ",".join(parameters)

    def _write_cloud_function_code(self, folder):

        task_folder = f"{folder}/{self.dag.dag_id}/{self.task_id}"
        shutil.rmtree(task_folder, ignore_errors=True)

        self.write_main(task_folder)
        self.write_requirements(task_folder)

        output_path = f"{task_folder}.zip"

        with zipfile.ZipFile(output_path, "w") as zip_file:
            self.create_zip_folder(task_folder, zip_file, os.path.abspath(task_folder))

        hash = hashlib.md5(open(f"{task_folder}.zip", "rb").read()).hexdigest()

        return hash

    def write_main(self, task_folder):
        code = inspect.getsource(self.python_callable)

        code += "\ndef cloudfunction_execution(event, context):\n"

        if self.provide_context:
            code += "    import base64\n" \
                    "    import json\n" \
                    "    if 'data' in event:\n"\
                    "        data = base64.b64decode(event['data']).decode('utf-8')\n"\
                    "        data = json.loads(data)\n"

        parameters = self._get_func_parameters(self.op_kwargs)

        code += f"    outputs = {self.python_callable.__name__}({parameters})\n"\
                f"\n    if outputs is None:\n"\
                f"        outputs = ['done']\n"\
                f"\n    topic_name = 'task_{self.dag.dag_id}_{self.task_id}'\n"\


        pubsub_code = inspect.getsourcelines(self._write_to_pub_sub_code)
        code += "\n    def call_pub_sub(message, topic_name):\n"
        code += "".join(pubsub_code[0][1:])

        code += "\n    for output in outputs:\n"
        code += "        call_pub_sub(output, topic_name)\n"

        main = f"{task_folder}/main.py"
        os.makedirs(os.path.dirname(main), exist_ok=True)

        with open(main, "w") as file:
            file.write(code)

    def write_requirements(self, task_folder):
        requirements = f"{task_folder}/requirements.txt"

        self.requirements.extend(["google-cloud", "google-cloud-pubsub"])

        with open(requirements, "w") as file:
            file.write("\n".join(self.requirements))

    def create_zip_folder(self, path, zip_file: zipfile, directory_root):
        for root, dirs, files in os.walk(path):
            for file in files:
                self.add_file(zip_file, os.path.join(root, file), os.path.relpath(f"{root}/{file}", directory_root))

    def add_file(self, zip_file: zipfile, file_path, zip_path=None):
        permission = 0o555 if os.access(file_path, os.X_OK) else 0o444
        zip_info = zipfile.ZipInfo.from_file(file_path, zip_path)
        zip_info.date_time = (2019, 1, 1, 0, 0, 0)
        zip_info.external_attr = (stat.S_IFREG | permission) << 16
        with open(file_path, "rb") as fp:
            zip_file.writestr(
                zip_info,
                fp.read(),
                compress_type=zipfile.ZIP_DEFLATED,
                compresslevel=9,
            )

    def _write_to_pub_sub_code(self, message, topic_name):
        from google.cloud import pubsub_v1
        import os
        import json

        PROJECT_ID = os.environ['GCP_PROJECT']
        publisher = pubsub_v1.PublisherClient()
        topic_path = publisher.topic_path(PROJECT_ID, topic_name)
        message_json = json.dumps({'data': {'message': message},})
        message_bytes = message_json.encode('utf-8')
        publish_future = publisher.publish(topic_path, data=message_bytes)
        publish_future.result()

    def get_terraform_json(self, *, folder) -> {}:

        hash = self._write_cloud_function_code(folder=folder)

        source_dir = f"{self.dag.dag_id}/{self.task_id}"

        if len(self.upstream_tasks) > 0:
            trigger_resource = f"task_{self.dag.dag_id}_{self.upstream_tasks[0]}"
        else:
            trigger_resource = f"dag_{self.dag.dag_id}"

        configuration = {
            "resource": {
                "google_storage_bucket_object": [{
                    f"task_{self.task_id}": {
                        "name": f"{source_dir}_{hash}.zip",
                        "bucket": "${google_storage_bucket.bucket.name}",
                        "source": f"{source_dir}.zip"
                    }
                }],
                "google_cloudfunctions_function": [{
                    f"function_{self.task_id}": {
                        "name": f"{self.dag.dag_id}-{self.task_id}",
                        "description": self.description,
                        "runtime": "python37",
                        "available_memory_mb": 256,
                        "service_account_email": os.environ["CLOUD_FUNCTIONS_SERVICE_ACCOUNT_EMAIL"],
                        "source_archive_bucket": "${google_storage_bucket.bucket.name}",
                        "source_archive_object": "${google_storage_bucket_object.task_"+self.task_id+".name}",
                        "event_trigger": {
                            "event_type": "providers/cloud.pubsub/eventTypes/topic.publish",
                            "resource": trigger_resource
                        },
                        "entry_point": "cloudfunction_execution"
                    }
                }],
                "google_pubsub_topic": [{
                    f"task_{self.dag.dag_id}_{self.task_id}": {
                        "name": f"task_{self.dag.dag_id}_{self.task_id}"
                    }
                }]
            }
        }

        return configuration


import os
import glob
import importlib
import sys
import hashlib
from ..models.dag import DAG
from ..models.operator import Operator
import json


class terraform:

    def __init__(self, folder):
        self.folder = folder

    def create_terraform_configuration(self):

        top_level_dags = self._get_objects(DAG)
        configuration = self._get_terraform_configuration_dags(top_level_dags)

        with open(f"{self.folder}/resources.tf.json","w") as file:
            file.write(json.dumps(configuration, indent=4))

    def _get_terraform_configuration_dags(self, dags):

        configurations = []

        for dag in dags:
            configurations.append(dag.get_terraform_json())

            tasks = self._get_objects(Operator)
            dag_tasks = [task for task in tasks if task.dag.dag_id == dag.dag_id]

            for task in dag_tasks:
                configurations.append(task.get_terraform_json(folder=self.folder))

        return self._combine_configurations(configurations)

    def _get_objects(self, object_type) -> []:

        files = glob.glob("dags/*.py")

        mods = []

        for filepath in files:
            org_mod_name, _ = os.path.splitext(os.path.split(filepath)[-1])
            path_hash = hashlib.sha1(filepath.encode('utf-8')).hexdigest()
            mod_name = f'unusual_prefix_{path_hash}_{org_mod_name}'

            loader = importlib.machinery.SourceFileLoader(mod_name, filepath)
            spec = importlib.util.spec_from_loader(mod_name, loader)
            new_module = importlib.util.module_from_spec(spec)
            sys.modules[spec.name] = new_module
            loader.exec_module(new_module)
            mods.append(new_module)

        top_level_dags = [
            o
            for m in mods
            for o in list(m.__dict__.values())
            if isinstance(o, object_type)
        ]

        return top_level_dags

    def _combine_configurations(self, configurations: []):

        terraform_configuration = {}

        # List of configurations
        for configuration in configurations:

            # Dict of configuration types (data or resource)
            for configuration_type in configuration:

                # If the configuration type is not yet in the config then add
                if configuration_type not in terraform_configuration:
                    terraform_configuration[configuration_type] = {}

                # List object types (pub_sub_topic, cloud_function)
                for object_type in configuration[configuration_type]:

                    # If object type not yet there then add
                    if object_type not in terraform_configuration[configuration_type]:
                        terraform_configuration[configuration_type][object_type] = {}

                    # List over objects of that type
                    for o in configuration[configuration_type][object_type]:

                        # List over keys of object (just 1 key anyway)
                        for key in o:
                            terraform_configuration[configuration_type][object_type][key] = o[key]

        return terraform_configuration




class Operator:

    def __init__(self):
        self.downstream_tasks = []
        self.upstream_tasks = []

    def __lshift__(self, task2):
        """Implements Task << Task"""
        self.set_upstream(task2)

    def __rshift__(self, task2):
        """Implements Task >> Task"""
        self.set_downstream(task2)

    def set_upstream(self, task):

        self.set_relative(task=task, upstream=True)
        task.set_relative(task=self, upstream=False)

    def set_downstream(self, task):

        self.set_relative(task=task, upstream=False)
        task.set_relative(task=self, upstream=True)

    def set_relative(self, *, task, upstream):

        if upstream:
            self.upstream_tasks.append(task.task_id)
        else:
            self.downstream_tasks.append(task.task_id)